"
 COPYRIGHT (c) 1989 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

AbsoluteTime subclass:#Time
       instanceVariableNames:''
       classVariableNames:''
       poolDictionaries:''
       category:'Magnitude-General'
!

Time comment:'
COPYRIGHT (c) 1989 by Claus Gittinger
	      All Rights Reserved

$Header: /cvs/stx/stx/libbasic/Time.st,v 1.12 1995-02-02 12:22:51 claus Exp $
'!

!Time class methodsFor:'documentation'!

copyright
"
 COPYRIGHT (c) 1989 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"
!

version
"
$Header: /cvs/stx/stx/libbasic/Time.st,v 1.12 1995-02-02 12:22:51 claus Exp $
"
!

documentation
"
    Instances of time represent a particular time-of-day.
"
! !

!Time class methodsFor:'instance creation'!

now
    "return an instance of Time representing this moment"

    ^ self fromOSTime:(OperatingSystem getTimeParts)

    "
     Time now
    "
!

fromOSTime:osTime
    "return a time, representing the time given by the operatingSystem time"

    ^ self basicNew setSecondsLow:(osTime at:1) and:(osTime at:2)
! !

!Time class methodsFor:'ST-80 compatibility'!

fromSeconds: seconds
    "return a Time that is constructed from seconds since midnight."

   ^ (self new setSeconds:seconds) subtractTime:60*60

    "midnight starts at 1:00:00 in ST/X for some strange reason"

    "
     Time fromSeconds:0
    "
!

hour:h minutes:m seconds:s
    "return an instance of Time representing the given values"

    | seconds |

    seconds := (h * 60 * 60 )+ (m * 60) + s.

    ^ self basicNew setSeconds:seconds.
!

totalSeconds
    ^ self secondClock
! !


!Time methodsFor:'accessing'!

asSeconds
    "return the number of seconds elapsed since midnight"

    OperatingSystem computeTimePartsOf:secondsLow and:secondsHi for:[
	:hours :minutes :seconds |

	^ ( 3600 * hours   )
	  + ( 60 * minutes )
	  + seconds
    ]

    "
     Time now asSeconds
    "
!

hours
    "return the number of hours since midnight (i.e. 0..23)"

    ^ self hourInDay

    "
     Time now hours
    "
!

minutes
    "return the number of minutes within the hour (i.e. 0..59)"

    ^self minuteInDay

    "
     Time now minutes
    "
!

seconds
    "return the number of seconds within the minute (i.e. 0..59)"

    ^self secondInDay

    "
     Time now seconds
    "
! !

!Time methodsFor:'comparing'!

> aTime
    "return true if the argument, aTime is after the receiver"

    (aTime respondsTo:#secondsLow) ifTrue:[
	(secondsHi > aTime secondsHi) ifTrue:[^ true].
	(secondsHi < aTime secondsHi) ifTrue:[^ false].
	(secondsLow > aTime secondsLow) ifTrue:[^ true].
	^ false
    ].
    ^ self getSeconds > aTime getSeconds
!

< aTime
    "return true if the argument, aTime is before the receiver"

    (aTime respondsTo:#secondsLow) ifTrue:[
	(secondsHi < aTime secondsHi) ifTrue:[^ true].
	(secondsHi > aTime secondsHi) ifTrue:[^ false].
	(secondsLow < aTime secondsLow) ifTrue:[^ true].
	^ false
    ].
    ^ self getSeconds < aTime getSeconds
! !

!Time methodsFor:'printing & storing'!

printOn:aStream
    "append a printed representation of the receiver to aStream.
     Format is hh:mm:ss either in 12-hour or 24-hour format.
     depending on the setting of LanguageTerritory.
     I dont know what ST-80 does here (12-hour format ?)"

    LanguageTerritory = 'usa' ifTrue:[
	self print12HourFormatOn:aStream
    ] ifFalse:[
	self print24HourFormatOn:aStream
    ]

    "
     Time now printOn:Transcript. Transcript cr
    "
!

print24HourFormatOn:aStream
    "append a printed representation of the receiver to aStream.
     Format is hh:mm:ss in 24-hour format."

    OperatingSystem computeTimePartsOf:secondsLow and:secondsHi for:[
	:hours :minutes :secs |

	hours printOn:aStream.
	aStream nextPut:$:.
	(minutes < 10) ifTrue:[aStream nextPut:$0].
	minutes printOn:aStream.
	aStream nextPut:$:.
	(secs < 10) ifTrue:[aStream nextPut:$0].
	secs printOn:aStream
    ]

    "
     Time now print24HourFormatOn:Transcript. Transcript cr
    "
!

print12HourFormatOn:aStream
    "append a printed representation of the receiver to aStream.
     Format is hh:mm:ss am/pm (i.e. 12-hour format)."

    OperatingSystem computeTimePartsOf:secondsLow and:secondsHi for:[
	:hours :minutes :secs | |h|

	h := hours.
	h > 12 ifTrue:[
	    h := h - 12.
	] ifFalse:[
	    h < 1 ifTrue:[
		h := 12
	    ]
	].
	h printOn:aStream.
	aStream nextPut:$:.
	(minutes < 10) ifTrue:[aStream nextPut:$0].
	minutes printOn:aStream.
	aStream nextPut:$:.
	(secs < 10) ifTrue:[aStream nextPut:$0].
	secs printOn:aStream.
	hours >= 12 ifTrue:[
	    aStream nextPutAll:' pm'
	] ifFalse:[
	    aStream nextPutAll:' am'
	]
    ]

    "
     Time now print12HourFormatOn:Transcript. Transcript cr
    "
! !
